"""
Convert python callables into MEL procedures
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future.utils import PY2
from builtins import filter
from builtins import zip
from builtins import range
from past.builtins import basestring
import inspect
import re
import types
from pymel.util.arguments import isMapping, isIterable
from pymel.core.language import getMelType, isValidMelType, MELTYPES
import pymel.api.plugins as plugins
import maya.mel as _mm
import maya.OpenMayaMPx as mpx
import maya.OpenMaya as om

if False:
    from typing import *

MAX_VAR_ARGS = 10
MAX_FLAG_ARGS = 6

_functionStore = {}


def _getFunction(function):
    # function is a string, so we must import its module and get the function object
    if isinstance(function, basestring):
        buf = function.split()
        funcName = buf.pop(-1)
        moduleName = '.'.join(buf)
        module = __import__(moduleName, globals(), locals(), [''])
        func = getattr(module, funcName)
    # function is a python object
    elif callable(function):
        func = function
    else:
        raise TypeError("argument must be a callable python object or the full, dotted path to the callable object as a string.")

    return func

# Flatten a multi-list argument so that in can be passed as
# a list of arguments to a command.


def getMelArgs(function, exactMelType=True):
    # type: (Union[Callable, str], Any) -> Tuple[Tuple[str, str], Dict[str, Any], Dict[str, str]]
    """Inspect the arguments of a python function and return the closest
    compatible MEL arguments.

    Parameters
    ----------
    function : Union[Callable, str]
        This can be a callable python object or the full, dotted path to the
        callable object as a string.

    Returns
    -------
    melArgs : Tuple[str, str]
        (argName, melType)
    melArgDefaults : Dict[str, Any]
        {argName : default}
    parsedDescr : Dict[str, str]
        {argName : description}
    """

    # FIXME: this whole function seems half implmented.  For instance,
    # it's only ever used once (AFAICT), and that usage doesn't make use of
    # either the returned melArgDefaults or the parsedDescr... and the
    # parsedDescr is currently always an empty dict.
    # Also, parsedTypes is supposed to somehow be used to detect arg types,
    # but it's never actually set to anything

    melArgs = []
    melArgDefaults = {}

    parsedTypes = {}
    parsedDescr = {}

    function = _getFunction(function)

    if PY2:
        args, varargs, kwargs, defaults = inspect.getargspec(function)
        if inspect.ismethod(function) and function.__self__ is not None:
            # remove self/cls
            args = args[1:]
        try:
            ndefaults = len(defaults)
        except:
            ndefaults = 0

        # print args, varargs, kwargs, defaults

        nargs = len(args)
        offset = nargs - ndefaults
        for i, arg in enumerate(args):

            if i >= offset:
                # keyword args with defaults
                default = defaults[i - offset]
                melType = getMelType(default, exactOnly=exactMelType)
                # a mel type of None means there is no mel analogue for this python object
                melArgDefaults[arg] = default
            else:
                # args without defaults
                # a mel type of None means there is no mel analogue for this python object
                melType = parsedTypes.get(arg, None)

            melArgs.append((arg, melType))
    else:
        parameters = inspect.signature(function).parameters

        validParamTypes = [inspect.Parameter.POSITIONAL_OR_KEYWORD,
                           inspect.Parameter.POSITIONAL_ONLY]

        # Note that in python3, the Signature for unbound methods will just be
        # a normal function (and therefore include self), while the Signature
        # for a bound method won't include self - so no special handling is
        # needed
        for arg, argInfo in parameters.items():
            if argInfo.kind not in validParamTypes:
                break
            if argInfo.default is not inspect.Parameter.empty:
                melType = getMelType(argInfo.default, exactOnly=exactMelType)
                # a mel type of None means there is no mel analogue for this python object
                melArgDefaults[arg] = argInfo.default
            else:
                # args without defaults
                # a mel type of None means there is no mel analogue for this python object
                melType = parsedTypes.get(arg, None)
            melArgs.append((arg, melType))

    return tuple(melArgs), melArgDefaults, parsedDescr


def py2melProc(function, returnType=None, procName=None, evaluateInputs=True, argTypes=None):
    # type: (Callable, Any, str, bool, Any) -> str
    """This is a work in progress.  It generates and sources a mel procedure which wraps the passed
    python function.  Theoretically useful for calling your python scripts in scenarios where Maya
    does not yet support python callbacks.

    The function is inspected in order to generate a MEL procedure which relays its
    arguments on to the python function.  However, Python features a very versatile argument structure whereas
    MEL does not.

        - python args with default values (keyword args) will be set to their MEL analogue, if it exists.
        - normal python args without default values default to strings. If 'evaluteInputs' is True, string arguments passed to the
            MEL wrapper proc will be evaluated as python code before being passed to your wrapped python
            function. This allows you to include a typecast in the string representing your arg::

                myWrapperProc( "Transform('persp')" );

        - *args : not yet implemented
        - **kwargs : not likely to be implemented

    Parameters
    ----------
    function : Callable
        This can be a callable python object or the full, dotted path to the callable object as a string.

        If passed as a python object, the object's __name__ and __module__ attribute must point to a valid module
        where __name__ can be found.

        If a string representing the python object is passed, it should include all packages and sub-modules, along
        with the function's name:  'path.to.myFunc'
    procName : str
        Optional name of the mel procedure to be created.  If None, the name of the function will be used.
    evaluateInputs : bool
        If True (default), string arguments passed to the generated mel procedure will be evaluated as python code, allowing
        you to pass a more complex python objects as an argument. For example:

        In python:
            >>> import pymel.tools.py2mel as py2mel
            >>> def myFunc( arg ):
            ...    for x in arg:
            ...       print(x)
            >>> py2mel.py2melProc( myFunc, procName='myFuncWrapper', evaluateInputs=True )

        Then, in mel::
            // execute the mel-to-python wrapper procedure
            myFuncWrapper("[ 1, 2, 3]");

        the string "[1,2,3]" will be converted to a python list [1,2,3] before it is executed by the python function myFunc

    Returns
    -------
    str
    """

    function = _getFunction(function)

    funcName = function.__name__

    melCompile = []
    melArgs = []
    melArrayToStrDecls = []
    argList, defaults, description = getMelArgs(function)

    if argTypes:
        if isMapping(argTypes):
            pass
        elif isIterable(argTypes):
            tmp = argTypes
            argTypes = {}
            for i, argType in enumerate(tmp):
                argTypes[argList[i][0]] = argType
        else:
            raise ValueError("argTypes must be iterable or mapping type")
        for argType in argTypes.values():
            if not isValidMelType(argType):
                raise TypeError("%r is not a valid mel type: %s" % (argType, ', '.join(MELTYPES)))
    else:
        argTypes = {}

    for arg, melType in argList:
        melType = argTypes.get(arg, melType)
        if melType == 'string':
            compilePart = "'\" + $%s + \"'" % arg
            melCompile.append(compilePart)
        elif melType == None:
            melType = 'string'
            compilePart = "'\" + $%s + \"'" % arg
            compilePart = r'eval(\"\"\"%s\"\"\")' % compilePart
            melCompile.append(compilePart)
        elif melType.count('[]'):
            melArrayToStrDecls.append('string $_%s ="("; int $i=0;for($i; $i<size($%s); $i++) { $_%s += ($%s[$i] + ",");  } $_%s += ")";' % (arg, arg, arg, arg, arg))
            melCompile.append("'\" + $_%s + \"'" % arg)
        else:
            melCompile.append("\" + $%s + \"" % arg)

        if melType.count('[]'):
            melType = melType.replace('[]', '')
            melArgs.append('%s $%s[]' % (melType, arg))
        else:
            melArgs.append('%s $%s' % (melType, arg))

    if procName is None:
        procName = funcName

    procDef = """global proc %s %s( %s ){ %s
    python("import %s; %s._functionStore[%r](%s)");}""" % (returnType if returnType else '',
                                                           procName,
                                                           ', '.join(melArgs),
                                                           ''.join(melArrayToStrDecls),
                                                           __name__,
                                                           __name__,
                                                           repr(function),
                                                           ','.join(melCompile))
    _functionStore[repr(function)] = function

    _mm.eval(procDef)
    return procName
#--------------------------------------------------------
#  Scripted Command Wrapper
#--------------------------------------------------------


def _shortnameByCaps(name):
    """
    uses hungarian notation (aka camelCaps) to generate a shortname, with a maximum of 3 letters
        ex.

            myProc --> mp
            fooBar --> fb
            superCrazyLongProc --> scl
    """

    shortname = name[0]
    count = 1
    for each in name[1:]:
        if each.isupper() or each.isdigit():
            shortname += each.lower()
            count += 1
            if count == 3:
                break
    return shortname


def _shortnameByUnderscores(name):
    """
    for python methods that use underscores instead of camelCaps, with a maximum of 3 letters
    """

    buf = name.split('_')
    shortname = ''
    for i, token in enumerate(buf):
        shortname += token[0].lower()
        if i == 2:
            break
    return shortname


def _shortnameByConvention(name):
    """
    chooses between byUnderscores and ByCaps
    """

    if '_' in name:
        return _shortnameByUnderscores(name)
    return _shortnameByCaps(name)


def _shortnameByDoc(method):
    """
    a shortname can be explicitly set by adding the keyword shortname followed by a colon followed by the shortname

            ex.

            class foo():
                def bar():
                    'shortname: b'
                    # do some things
                    return

    """
    if hasattr(method, "__doc__") and method.__doc__:
        m = re.search(r'.*shortname: (\w+)', method.__doc__)
        if m:
            return m.group(1)


def _nonUniqueName(longname, shortname, shortNames, operation):
    if operation in ['skip', 'warn', 'error'] and shortname in shortNames:
        message = "default short name %r for flag %r is taken" % (shortname, longname)
        if operation == 'warn':
            print('warning: ' + message)
            return False
        elif operation == 'skip':
            print('skipping: ' + message)
            return True
        else:
            raise TypeError(message)


def _invalidName(commandName, longname, operation):
    if len(longname) < 4 and operation in ['skip', 'warn', 'error']:
        message = 'long flag names must be at least 4 characters long: %s -%r' % (commandName, longname.lower())
        if operation == 'warn':
            print('warning: ' + message)
            return False
        elif operation == 'skip':
            print('skipping: ' + message)
            return True
        else:
            raise TypeError(message)


def _getShortNames(objects, nonUniqueName):
    """uses several different methods to generate a shortname flag from the long name"""
    shortNames = []
    nonunique = {}
    for obj in objects:
        if isinstance(obj, (list, tuple)):
            longname = obj[0]
            if obj[0] is not None:
                shortname = _shortnameByDoc(obj[1])
        else:
            longname = obj
            shortname = None
        # try _shortnameByDoc first

        if not shortname:
            shortname = _shortnameByConvention(longname)

        if not shortname or shortname in shortNames:
            if _nonUniqueName(longname, shortname, shortNames, nonUniqueName):
                shortname = None
            else:
                shortname = longname[0]
                unique = False
                for each in longname[1:3]:
                    shortname += each.lower()
                    if shortname not in shortNames:
                        unique = True
                        break
                if not unique:
                    baseshort = shortname[:2]
                    while True:
                        count = nonunique.get(baseshort, 0) + 1
                        nonunique[baseshort] = count
                        shortname = baseshort + str(count)
                        if shortname not in shortNames:
                            break
                    # print 'could not find a unique shortname for %s: using %s'% ( methodName, shortname )
        shortNames.append((longname, shortname))
    return tuple(shortNames)


def _getArgInfo(obj, allowExtraKwargs=True, maxVarArgs=MAX_VAR_ARGS,
                filter=None, method=False):
    '''Returns a dict giving info about the arugments for the function/property

    If obj is None, will return the 'defaults'.
    '''
    # per flag query and edit settings
    canQuery = False
    canEdit = False
    defaults = {}
    argNames = []

    # object.__init__ is a 'slot wrapper', but can't find that type defined
    # anywhere...
    if isinstance(obj, (types.BuiltinFunctionType, type(object.__init__),
                        type(None))):
        maxArgs = 0
    elif isinstance(obj, property):
        # enable edit and query to determine what the user intends to do with this get/set property
        if obj.fget:
            canQuery = True
        if obj.fset:
            canEdit = True
        maxArgs = 1
    else:
        if PY2:
            argNames, extraArgs, extraKwargs, defaults = inspect.getargspec(obj)
            if defaults is None:
                defaults = {}

            # turn defaults into a dict
            defaults = dict(zip(argNames[-len(defaults):], defaults))
        else:
            defaults = {}
            parameters = inspect.signature(obj).parameters
            extraArgs = extraKwargs = False
            for argName, argInfo in parameters.items():
                argNames.append(argName)
                if argInfo.kind == argInfo.VAR_POSITIONAL:
                    extraArgs = True
                elif argInfo.kind == argInfo.VAR_KEYWORD:
                    extraKwargs = True
                elif argInfo.kind == argInfo.KEYWORD_ONLY or (
                        argInfo.kind == argInfo.POSITIONAL_OR_KEYWORD
                        and argInfo.default is not argInfo.empty):
                    # Note that it's possible to have KEYWORD_ONLY args, that
                    # have no default; we still add these to defaults, as
                    # the dict is currently only used to determine whether
                    # the args should be treated as args or flags (the
                    # default vaules aren't used).
                    defaults[argName] = argInfo.default

        if method:
            # remove the self arg
            del argNames[0]

        if extraKwargs and not allowExtraKwargs:
            raise ValueError('arguments of the format **kwargs are not supported')

        if filter:
            newArgNames = []
            for argName in argNames:
                # can only filter out optional args
                if argName not in defaults or filter(argName):
                    newArgNames.append(argName)
                else:
                    del defaults[argName]
            argNames = newArgNames

        # a variable number of args can be passed to the flag. set the maximum number
        if extraArgs:
            maxArgs = maxVarArgs
        else:
            maxArgs = len(argNames)

    return {'maxArgs': maxArgs, 'canQuery': canQuery, 'canEdit': canEdit,
            'argNames': argNames, 'defaults': defaults}


class WrapperCommand(plugins.Command):
    _syntax = None
    _flagInfo = None
    _mainArgInfo = None

    @classmethod
    def createSyntax(cls):
        return cls._syntax

    def setResult(self, result):
        """
        convert to a valid result type
        """

#        int
#        double
#        bool
#        const MString
#        const MIntArray
#        const MDoubleArray
#        const MStringArray

        if result is None:
            return

        if isinstance(result, dict):
            # convert a dictionary into a 2d list
            newResult = []
            for key, value in list(result.items()):
                newResult.append(key)
                newResult.append(value)
            mpx.MPxCommand.setResult(newResult)
        else:
            # try:
            mpx.MPxCommand.setResult(result)

    def parseCommandArgs(self, argData):
        argValues = []
        i = 0
        while(1):
            try:
                argValues.append(argData.commandArgumentString(i).encode())
            except RuntimeError:
                break
            else:
                i += 1
        return argValues

    def parseFlagArgs(self, argData):
        # type: (Any) -> List[Tuple[str, List[str]]]
        """
        cycle through known flags looking for any that have been set.

        Returns
        -------
        List[Tuple[str, List[str]]]
            a list of (flagLongName, flagArgList) tuples
        """

        argValues = []
        for flag in self._flagInfo:
            if argData.isFlagSet(flag):
                canQuery = self._flagInfo[flag].get('canQuery', False)
                canEdit = self._flagInfo[flag].get('canEdit', False)
                if argData.isQuery():
                    if not canQuery:
                        raise SyntaxError('cannot use the query flag with %s' % flag)
                elif argData.isEdit():
                    if not canEdit:
                        raise SyntaxError('cannot use the query edit with %s' % flag)
                elif canQuery or canEdit:
                    raise SyntaxError('the %s flag must be used with either query or edit' % flag)

                flagArgs = []
                maxArgs = self._flagInfo[flag]['maxArgs']
                for i in range(maxArgs):
                    try:
                        flagArgs.append(argData.flagArgumentString(flag, i))
                    except Exception:
                        break

                argValues.append((flag, flagArgs))
        return argValues


def py2melCmd(pyObj, commandName=None, register=True, includeFlags=None,
              excludeFlags=[], includeFlagArgs=None, excludeFlagArgs={},
              nonUniqueName='warn', invalidName='warn'):
    # type: (Any, str, bool, List[str], List[str], Dict[str, List[str]], Dict[str, List[str]], str, str) -> None
    """
    Create a MEL command from a python function or class.

    A MEL command has two types of arguments: command arguments and flag arguments.  In the case of passing a function, the function's
    non-keyword arguments become the command arguments and the function's keyword arguments become the flag arguments.
    for example::

        def makeName( first, last, middle=''):
            if middle:
                return first + ' ' + middle + ' ' + last
            return first + ' ' + last

        import pymel as pm
        from pymel.tools.py2mel import py2melCmd
        cmd = py2melCmd( makeName, 'makeNameCmd' )
        pm.makeNameCmd( 'Homer', 'Simpson')
        # Result: Homer Simpson #
        pm.makeNameCmd( 'Homer', 'Simpson', middle='J.')
        # Result: Homer J. Simpson #

    Of course, the real advantage of this tool is that now your python function is available from within MEL as a command::

        makeNameCmd "Homer" "Simpson";
        // Result: Homer Simpson //
        makeNameCmd "Homer" "Simpson" -middle "J.";
        // Result: Homer J. Simpson //

    To remove the command, call the deregister method of the class returned by py2melCmd::

        cmd.deregister()

    This function attempts to automatically create short names (3 character max) based on the long names of the methods or arguments of the pass python object.
    It does this by looping through long names in alphabetical order and trying the following techniques until a unique short name is found:

            1. by docstring (methods only): check the method docstring looking for something of the form ``shortname: xyz``::
                class Foo():
                    def bar():
                        'shortname: b'
                        # do some things
                        return
            2. by convention:  if the name uses under_scores or camelCase, use the first letter of each "word" to generate a short name up to 3 letters long
            3. first letter
            4. first two letters
            5. first three letters
            6. first two letters plus a unique digit

    .. warning:: if you edit the python object that is passed to this function it may result in short names changing!  for example, if you have a class like the following::

                class Foo():
                    def bar():
                        pass

            ``Foo.bar`` will be assigned the short flag name 'b'. but if you later add the method ``Foo.baa``, it will be assigned the short flag name 'b' and 'bar' will be given 'ba'.
            **The only way to be completely sure which short name is assigned is to use the docstring method described above.**

    Parameters
    ----------
    commandName : str
        name given to the generated MEL command
    register : bool
        whether or not to automatically register the generated command.  If
        False, you will have to manually call the `register` method of the
        returned `WrapperCommand` instance
    includeFlags : List[str]
        list of flags to include. if given, other flags will be ignored
    excludeFlags : List[str]
        list of flags to exclude
    includeFlagArgs : Dict[str, List[str]]
        for each flag, a list of arg names to include; if given, other args will
        be ignored
    excludeFlagArgs : Dict[str, List[str]]
        for each flag, a list of arg names to exclude
    nonUniqueName : str
        {'force', 'warn', 'skip', or 'error'}
        what to do if a flag name is not unique
    invalidName: str
        {'force', 'warn', 'skip', or 'error'}
        what to do if a flag name is invalid

    """
    if not commandName:
        commandName = pyObj.__name__
    if includeFlagArgs is None:
        includeFlagArgs = {}

    def goodFlag(flag):
        return ((includeFlags is None or flag in includeFlags)
                and flag not in excludeFlags and not flag.startswith('_'))

    syntax = om.MSyntax()
    flagInfo = {}
    flags = []   # ordered list of flags
    if inspect.isfunction(pyObj):
        # args         --> command args
        # keyword args --> flags
        classWrap = False

        mainArgInfo = _getArgInfo(pyObj, allowExtraKwargs=False,
                                  filter=goodFlag)
        ndefaults = len(mainArgInfo['defaults'])

        # positional args become the command args
        cmdArgs = mainArgInfo['argNames'][:-ndefaults]

        # keyword args become the flags
        flags = mainArgInfo['argNames'][-ndefaults:]
        flags = [x for x in flags if goodFlag(x)]
        mainArgInfo['maxArgs'] = len(cmdArgs)
        mainArgInfo['argNames'] = cmdArgs

        for flag in flags:
            # currently keyword args only support one item per flag. eventually we may
            # detect when a keyword expects a list as an argument
            flagInfo[flag] = {'maxArgs': 1}

    elif inspect.isclass(pyObj):
        # init/new args--> command args
        # methods      --> flags
        # method args  --> flag args
        classWrap = True

        # __init__ or __new__ becomes the command args
        try:
            initFunc = pyObj.__init__
        except AttributeError:
            initFunc = None
        if initFunc is None or initFunc == object.__init__:
            initFunc = object.__new__
        mainArgInfo = _getArgInfo(initFunc, filter=goodFlag, method=True)

        # methods / properties become the flag args
        def isFlagCovertible(x):
            return inspect.isfunction(x) or isinstance(x, property)

        if PY2:
            def isFlagCovertible(x):
                return inspect.ismethod(x) or isinstance(x, property)

        for longname, method in inspect.getmembers(pyObj, isFlagCovertible):
            if not goodFlag(longname):
                continue
            flags.append(longname)

            includeArgs = includeFlagArgs.get(longname)
            excludeArgs = excludeFlagArgs.get(longname, ())

            def goodFlagArg(flagArg):
                return ((includeArgs is None or flagArg in includeArgs)
                        and flagArg not in excludeArgs)

            argInfo = _getArgInfo(method, maxVarArgs=MAX_FLAG_ARGS,
                                  filter=goodFlagArg, method=True)
            argInfo['method'] = method
            argInfo['methodName'] = longname
            argInfo['type'] = type(method)
            flagInfo[longname] = argInfo

    # command args
    for _ in range(mainArgInfo['maxArgs']):
        syntax.addArg(om.MSyntax.kString)

    for origname, shortname in _getShortNames(flags, nonUniqueName):
        if shortname is None:
            # we skipped the flag
            continue
        longname = origname
        if _invalidName(commandName, longname, invalidName):
            continue
        if len(longname) < 4:
            longname = longname.ljust(4, 'x')

        # TODO: currently no check that LONG name is unique...
        # ...maybe roll this into _getShortNames... make it _getNames?
        if longname != origname:
            flagInfo[longname] = flagInfo.pop(origname)

        thisFlagInfo = flagInfo[longname]
        # NOTE: shortname and longname MUST be stored on the class or they will
        # get garbage collected and the names will be destroyed
        thisFlagInfo['shortname'] = shortname

        if thisFlagInfo['canQuery']:
            syntax.enableQuery(True)
        if thisFlagInfo['canEdit']:
            syntax.enableEdit(True)

        # currently keyword args only support one item per flag. eventually we may
        # detect when a keyword expects a list as an argument
        syntaxArgs = [shortname, longname] + ([om.MSyntax.kString] * thisFlagInfo['maxArgs'])
        syntax.addFlag(*syntaxArgs)

    class dummyCommand(WrapperCommand):
        _syntax = syntax
        _flagInfo = flagInfo
        _mainArgInfo = mainArgInfo

        def doIt(self, argList):

            argData = om.MArgParser(self.syntax(), argList)

            cmdArgs = self.parseCommandArgs(argData)
            flagArgs = self.parseFlagArgs(argData)

            if not classWrap:
                # doing a function wrap...

                # unpack the flag arguments, there should always only be 1
                kwargs = dict([(x[0], x[1][0]) for x in flagArgs])

                res = pyObj(*cmdArgs, **kwargs)
            else:
                # doing a class wrap...

                if len(flagArgs) != 1:
                    raise RuntimeError('only one flag can be used at a time for command %s' % commandName)
                longname, flagArgs = flagArgs[0]

                inst = pyObj(*cmdArgs)
                flagInfo = self._flagInfo[longname]
                attrType = flagInfo['type']
                methodName = flagInfo['methodName']

                # PY2: once python-3 only, can probably remove the
                # types.MethodType check
                if issubclass(attrType, (types.FunctionType, types.MethodType)):  # method
                    # build out the args and kwargs...
                    # ...can't just pass flagArgs straight into our method using
                    #   myMethod(*flagArgs)
                    # ...because we may have filtered out some optional args, so
                    # the positional information isn't right...
                    ndefaults = len(flagInfo['defaults'])
                    args = flagArgs[:-ndefaults]

                    defaultNames = flagInfo['argNames'][-ndefaults:]
                    defaultVals = flagArgs[-ndefaults:]
                    kwargs = dict(list(zip(defaultNames, defaultVals)))

                    res = getattr(inst, methodName)(*args, **kwargs)
                else:  # property
                    if argData.isQuery():
                        res = getattr(inst, methodName)
                    elif argData.isEdit():
                        # property-defined flags can only take one arg
                        if len(flagArgs) != 1:
                            raise RuntimeError('flag %s for command %s may only have one arg' % (longname, commandName))
                        res = setattr(inst, methodName, flagArgs[0])
                    else:
                        raise SyntaxError("properties must either be edited or queried")
            return self.setResult(res)

    dummyCommand.__name__ = commandName
    if register:
        dummyCommand.register()
    return dummyCommand




"""
Defines common types and type related utilities:  Singleton, etc.
These types can be shared by other utils modules and imported into util main namespace for use by other pymel modules
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from future.utils import PY2, with_metaclass

# 2to3: remove switch when python-3 only
if PY2:
    from collections import Mapping
else:
    from collections.abc import Mapping
from builtins import object
import inspect
import types
import operator
import sys
import warnings
from collections import defaultdict

if False:
    from typing import Any, Iterable, Type, TypeVar
    T = TypeVar('T')


class Singleton(type):

    """ Metaclass for Singleton classes.

        >>> class DictSingleton(with_metaclass(Singleton, dict)) :
        ...     pass
        ...
        >>> DictSingleton({'A':1})
        {'A': 1}
        >>> a = DictSingleton()
        >>> a
        {'A': 1}
        >>> b = DictSingleton({'B':2})
        >>> a, b, DictSingleton()
        ({'B': 2}, {'B': 2}, {'B': 2})
        >>> a is b and a is DictSingleton()
        True

        >>> class StringSingleton(with_metaclass(Singleton, str)) :
        ...    pass
        ...
        >>> StringSingleton("first")
        'first'
        >>> a = StringSingleton()
        >>> a
        'first'
        >>> b = StringSingleton("changed")
        >>> a, b, StringSingleton()
        ('first', 'first', 'first')
        >>> a is b and a is StringSingleton()
        True

        >>> class DictSingleton2(DictSingleton):
        ...     pass
        ...
        >>> DictSingleton2({'A':1})
        {'A': 1}
        >>> a = DictSingleton2()
        >>> a
        {'A': 1}
        >>> b = DictSingleton2({'B':2})
        >>> a, b, DictSingleton2()
        ({'B': 2}, {'B': 2}, {'B': 2})
        >>> a is b and a is DictSingleton2()
        True

    """
    def __new__(mcl, classname, bases, classdict):

        # redefine __new__
        def __new__(cls, *p, **k):
            if '_the_instance' not in cls.__dict__:
                cls._the_instance = super(newcls, cls).__new__(cls, *p, **k)
            self = cls._the_instance
            if p:
                if hasattr(self, 'clear'):
                    self.clear()
                else:
                    self.__init__()
            return self
        newdict = {'__new__': __new__}
        for k in classdict:
            if k in newdict:
                warnings.warn("Attribute %r is predefined in class %r of "
                              "type %r and can't be overriden" %
                              (k, classname, mcl.__name__))
            else:
                newdict[k] = classdict[k]

        newcls = super(Singleton, mcl).__new__(mcl, classname, bases, newdict)

        return newcls


class metaStatic(Singleton):

    """ A static (immutable) Singleton metaclass to quickly build classes
        holding predefined immutable dicts

        >>> class FrozenDictSingleton(with_metaclass(metaStatic, dict)) :
        ...    pass
        ...
        >>> FrozenDictSingleton({'A':1})
        {'A': 1}
        >>> a = FrozenDictSingleton()
        >>> a
        {'A': 1}
        >>> b = FrozenDictSingleton()
        >>> a, b
        ({'A': 1}, {'A': 1})
        >>> a is b
        True

        >>> b = FrozenDictSingleton({'B':2})
        Traceback (most recent call last):
            ...
        TypeError: 'FrozenDictSingleton' object does not support redefinition

        >>> a['A']
        1
        >>> a['A'] = 2   #doctest: +ELLIPSIS
        Traceback (most recent call last):
            ...
        TypeError: '<class '...FrozenDictSingleton'>' object does not support item assignation

        >>> a.clear()
        Traceback (most recent call last):
            ...
        AttributeError: 'FrozenDictSingleton' object has no attribute 'clear'

        >>> a, b, FrozenDictSingleton()
        ({'A': 1}, {'A': 1}, {'A': 1})
        >>> a is b and a is FrozenDictSingleton()
        True

        >>> class StaticTest(FrozenDictSingleton):
        ...     pass
        ...
        >>> StaticTest({'A': 1})
        {'A': 1}
        >>> a = StaticTest()
        >>> a
        {'A': 1}
        >>> b = StaticTest()
        >>> a, b
        ({'A': 1}, {'A': 1})

        >>> class StaticTest2( StaticTest ):
        ...     pass
        ...
        >>> StaticTest2({'B': 2})
        {'B': 2}
        >>> a = StaticTest2()
        >>> a
        {'B': 2}
        >>> b = StaticTest2()
        >>> a, b
        ({'B': 2}, {'B': 2})

    """
    def __new__(mcl, classname, bases, classdict):
        """
        """
        # redefine __init__

        def __init__(self, *p, **k):
            cls = self.__class__
            # Can only create once)
            if p:
                # Can only init once
                if not self:
                    return super(newcls, self).__init__(*p, **k)
                else:
                    raise TypeError("'" + classname + "' object does not support redefinition")
        newdict = {'__init__': __init__}
        # hide methods with might herit from a mutable base

        def __getattribute__(self, name):
            if name in newcls._hide:
                raise AttributeError("'" + classname + "' object has no attribute '" + name + "'")
            else:
                return super(newcls, self).__getattribute__(name)
        newdict['__getattribute__'] = __getattribute__
        _hide = ('clear', 'update', 'pop', 'popitem', '__setitem__', '__delitem__', 'append', 'extend')
        newdict['_hide'] = _hide
        # prevent item assignation or deletion

        def __setitem__(self, key, value):
            raise TypeError("'%s' object does not support item assignation" % (self.__class__))
        newdict['__setitem__'] = __setitem__

        def __delitem__(self, key):
            raise TypeError("'%s' object does not support item deletion" % (self.__class__))
        newdict['__delitem__'] = __delitem__
        # Now add methods of the defined class, as long as it doesn't try to redefine
        # Note: could have defined the __new__ method like it is done in Singleton but it's as easy to derive from it
        for k in classdict:
            if k in newdict:
                warnings.warn("Attribute %r is predefined in class %r of type %r and can't be overriden" % (k, classname, mcl.__name__))
            else:
                newdict[k] = classdict[k]

        newcls = super(metaStatic, mcl).__new__(mcl, classname, bases, newdict)

        return newcls


class defaultlist(list):

    def __init__(self, default_factory, *args, **kwargs):
        if (default_factory is not None and
                not hasattr(default_factory, '__call__')):
            raise TypeError('first argument must be callable')
        list.__init__(self, *args, **kwargs)
        self.default_factory = default_factory

    def __setitem__(self, index, item):
        try:
            list.__setitem__(self, index, item)
        except IndexError:
            diff = index - len(self) - 1
            assert diff > 0
            self.extend([self.default_factory()] * diff + [item])

    def __getitem__(self, index):
        try:
            return list.__getitem__(self, index)
        except IndexError:
            return self.default_factory()


class ModuleInterceptor(object):

    """
    This class is used to intercept an unset attribute of a module to perfrom a callback. The
    callback will only be performed if the attribute does not exist on the module. Any error raised
    in the callback will cause the original AttributeError to be raised.

        def cb( module, attr):
             if attr == 'this':
                 print "intercepted"
             else:
                 raise ValueError
        import sys
        sys.modules[__name__] = ModuleInterceptor(__name__, cb)
        intercepted

    The class does not work when imported into the main namespace.
    """

    def __init__(self, moduleName, callback):
        self.module = __import__(moduleName, globals(), locals(), [''])
        self.callback = callback

    def __getattr__(self, attr):
        try:
            return getattr(self.module, attr)
        except AttributeError as msg:
            try:
                self.callback(self.module, attr)
            except:
                raise AttributeError(msg)

# read only decorator


def readonly(f):
    """ Marks a class member as protected, allowing metaProtected to prevent
    re-assignation on the classes it generates
    """
    f.__readonly__ = None
    return f


class metaReadOnlyAttr(type):

    """ A metaclass to allow to define read-only class attributes, accessible either on the class or it's instances
        and protected against re-write or re-definition.
        Read only attributes are stored in the class '__readonly__' dictionary.
        Any attribute can be marked as read only by including its name in a tuple named '__readonly__' in the class
        definition. Alternatively methods can be marked as read only with the @readonly decorator and will then get
        added to the dictionary at class creation """

    def __setattr__(cls, name, value):  # @NoSelf
        """ overload __setattr__ to forbid modification of read only class info """
        readonly = {}
        for c in inspect.getmro(cls):
            if hasattr(c, '__readonly__'):
                readonly.update(c.__readonly__)
        if name in readonly:
            raise AttributeError("attribute %s is a read only class attribute and cannot be modified on class %s" % (name, cls.__name__))
        else:
            super(metaReadOnlyAttr, cls).__setattr__(name, value)

    def __new__(mcl, classname, bases, classdict):  # @NoSelf
        """ Create a new metaReadOnlyAttr class """

        # checks for protected members, in base classes on in class to be created
        readonly = {}
        # check for protected members in class definition
        if '__readonly__' in classdict:
            readonly.update(dict((a, None) for a in classdict['__readonly__']))
        for a in classdict:
            if hasattr(classdict[a], '__readonly__'):
                readonly[a] = None
        readonly['__readonly__'] = None
        classdict['__readonly__'] = readonly

        # create the new class
        newcls = super(metaReadOnlyAttr, mcl).__new__(mcl, classname, bases, classdict)

        return newcls

NOT_PROXY_WRAPPED = ['__new__', '__getattribute__', '__getattr__', '__setattr__',
                     '__class__', '__weakref__', '__subclasshook__',
                     '__reduce_ex__', '__reduce__', '__dict__', '__sizeof__',
                     '__module__', '__init__', '__doc__']


def proxyClass(cls, classname, dataAttrName=None, dataFuncName=None,
               remove=(), makeDefaultInit=False, sourceIsImmutable=True,
               module=None):
    # type: (Type[T], str, str, str, Iterable[str], bool, bool, Any) -> Type[T]
    """
    This function will generate a proxy class which keeps the internal data
    separate from the wrapped class. This is useful for emulating immutable
    types such as str and tuple, while using mutable data.  Be aware that
    changing data will break hashing.  not sure the best solution to this,
    but a good approach would be to subclass your proxy and implement
    a valid __hash__ method.

    Parameters
    ----------
    cls : Type[T]
        The class to wrap
    classname : str
        The name to give the resulting proxy class
    dataAttrName : str
        The name of an attribute on which an instance of the wrapped class will
        be stored.
        Either dataAttrname or dataFuncName must be given, but not both.
    dataFuncName : str
        The name of an attribute on which reside a function, which takes no
        arguments, and when called, will return an instance of the wrapped
        class.
        Either dataAttrname or dataFuncName must be given, but not both.
    remove : Iterable[str]
        An iterable of name of attributes which should NOT be wrapped.
        Note that certain attributes will never be wrapped - the list of
        such items is found in the NOT_PROXY_WRAPPED constant.
    makeDefaultInit : bool
        If True and dataAttrName is True, then a 'default' __init__ function
        will be created, which creates an instance of the wrapped class, and
        assigns it to the dataAttr. Defaults to False
        If dataAttrName is False, does nothing
    sourceIsImmutable : bool
        This parameter is included only for backwards compatibility - it is
        ignored.

    Returns
    -------
    TypeT
    """

    assert not (dataAttrName and dataFuncName), \
        'Cannot use attribute and function for data storage. Choose one or the other.'

    if dataAttrName:
        class ProxyAttribute(object):

            def __init__(self, name):
                self.name = name

            def __get__(self, proxyInst, proxyClass):
                if proxyInst is None:
                    return getattr(cls, self.name)
                else:
                    return getattr(getattr(proxyInst, dataAttrName),
                                   self.name)

        def _methodWrapper(method):
            def wrapper(self, *args, **kwargs):
                return method(getattr(self, dataAttrName), *args, **kwargs)

            wrapper.__doc__ = method.__doc__
            wrapper.__name__ = method.__name__
            return wrapper

    elif dataFuncName:
        class ProxyAttribute(object):

            def __init__(self, name):
                self.name = name

            def __get__(self, proxyInst, proxyClass):
                if proxyInst is None:
                    return getattr(cls, self.name)
                else:
                    return getattr(getattr(proxyInst, dataFuncName)(),
                                   self.name)

        def _methodWrapper(method):
            def wrapper(self, *args, **kwargs):
                return method(getattr(self, dataFuncName)(), *args, **kwargs)

            wrapper.__doc__ = method.__doc__
            wrapper.__name__ = method.__name__
            return wrapper
    else:
        raise TypeError('Must specify either a dataAttrName or a dataFuncName')

    class Proxy(object):
        # make a default __init__ which sets the dataAttr...
        # if __init__ is in remove, or dataFuncName given,
        # user must supply own __init__, and set the dataAttr/dataFunc
        # themselves
        if makeDefaultInit and dataAttrName:
            def __init__(self, *args, **kwargs):
                # We may wrap __setattr__, so don't use 'our' __setattr__!
                object.__setattr__(self, dataAttrName, cls(*args, **kwargs))

        # For 'type' objects, you can't set the __doc__ outside of
        # the class definition, so do it here:
        if '__doc__' not in remove:
            __doc__ = cls.__doc__

    remove = set(remove)
    remove.update(NOT_PROXY_WRAPPED)
    for attrName, attrValue in inspect.getmembers(cls):
        if attrName not in remove:
            # We wrap methods using _methodWrapper, because if someone does
            #    unboundMethod = MyProxyClass.method
            # ...they should be able to call unboundMethod with an instance
            # of MyProxyClass as they expect (as opposed to an instance of
            # the wrapped class, which is what you would need to do if
            # we used ProxyAttribute)

            # ...the stuff with the cls.__dict__ is just to check
            # we don't have a classmethod - since it's a data descriptor,
            # we have to go through the class dict...

            # PY2: once we completely transition to PY3 only, can probably
            # remove the ismethod check
            if ((inspect.ismethoddescriptor(attrValue) or
                 inspect.isfunction(attrValue) or
                 inspect.ismethod(attrValue)) and
                not isinstance(cls.__dict__.get(attrName, None),
                               (classmethod, staticmethod))):
                try:
                    setattr(Proxy, attrName, _methodWrapper(attrValue))
                except AttributeError:
                    print("proxyClass: error adding proxy method %s.%s" % (classname, attrName))
            else:
                try:
                    setattr(Proxy, attrName, ProxyAttribute(attrName))
                except AttributeError:
                    print("proxyClass: error adding proxy attribute %s.%s" % (classname, attrName))

    Proxy.__name__ = classname
    if module is not None:
        Proxy.__module__ = module
    return Proxy


# Note - for backwards compatibility reasons, PyNodes still inherit from
# ProxyUnicode, even though we are now discouraging their use 'like strings',
# and ProxyUnicode itself has now had so many methods removed from it that
# it's no longer really a good proxy for unicode.

# NOTE: This may move back to core.general, depending on whether the
# __getitem__ bug was fixed in 2009, since we'll have to do a version switch there
# ProxyUnicode = proxyClass( unicode, 'ProxyUnicode', dataFuncName='name',
#                           remove=['__getitem__', 'translate'])
# 2009 Beta 2.1 has issues with passing classes with __getitem__
if PY2:
    _proxyStrBase = unicode
else:
    _proxyStrBase = str
ProxyUnicode = proxyClass(
    _proxyStrBase, 'ProxyUnicode',
    module=__name__, dataFuncName='name',
    remove=['__doc__', '__getslice__', '__contains__', '__len__',
            '__mod__', '__rmod__', '__mul__', '__rmod__', '__rmul__',  # reserved for higher levels
            'expandtabs', 'translate', 'decode', 'encode', 'splitlines',
            'capitalize', 'swapcase', 'title',
            'isalnum', 'isalpha', 'isdigit', 'isspace', 'istitle',
            'zfill'])


class universalmethod(object):
    #    """
    #    a decorator which is similar to builtin classmethod, but which leaves the method unmodified when called
    #    as a normal instance method:
    #        - when the wrapped method is called as a class method, the first argument will be the class.
    #        - when the wrapped method is called as an instance method, the first argument will be the instance.
    #
    #        >>> import inspect
    #        >>> class D(object):
    #        ...    @universalmethod
    #        ...    def f( obj ):
    #        ...        if inspect.isclass(obj):
    #        ...            print "doing something class related"
    #        ...        else:
    #        ...            print "doing something instance related"
    #        ...
    #        >>> D.f()
    #        doing something class related
    #        >>> d = D()
    #        >>> d.f()
    #        doing something instance related
    #
    #    """

    def __init__(self, f):
        self.f = f
        self.__doc__ = f.__doc__

    def __get__(self, instance, cls=None):
        if cls is None:
            cls = type(instance)
        if instance is None:
            instance = cls

        def newfunc(*args, **kwargs):
            return self.f(instance, *args, **kwargs)
        newfunc.__doc__ = self.__doc__
        return newfunc


class LazyLoadModule(types.ModuleType):

    """
    :param name: name of the module
    :param contents: dictionary of initial module globals

    This function returns a special module type with one method `_lazyModule_addAttr`.  The signature
    of this method is:

        _lazyModule_addAttr(name, creator, *creatorArgs, **creatorKwargs)

    Attributes added with this method will not be created until the first time that
    they are accessed, at which point a callback function will be called to generate
    the attribute's value.

    :param name: name of the attribute to lazily add
    :param creator: a function that create the

    Example::

        import sys
        mod = LazyLoadModule(__name__, globals())
        mod._lazyModule_addAttr( 'foo', str, 'bar' )

    One caveat of this technique is that if a user imports everything from your
    lazy module ( .e.g from module import * ), it will cause all lazy attributes
    to be evaluated.

    Also, if any module-level expression needs to reference something that only
    exists in the LazyLoadModule, it will need to be stuck in after the creation of the
    LazyLoadModule.  Then, typically, after defining all functions/classes/etc
    which rely on the LazyLoadModule attributes, you will wish to update the
    LazyLoadModule with the newly-created functions - typically, this is done
    with the _updateLazyModule method.

    Finally, any functions which reference any LazyLoadModule-only attributes,
    whether they are defined after OR before the creation of the LazyLoadModule,
    will have to prefix it with a reference to the LazyLoadModule.

    Example::

        import sys

        def myFunc():
            # need to preface foo with 'lazyModule',
            # even though this function is defined before
            # the creation of the lazy module!
            print 'foo is:', lazyModule.foo

        lazyModule = LazyLoadModule(__name__, globals())
        lazyModule._lazyModule_addAttr( 'foo', str, 'bar' )

        # define something which relies on something in the lazy module
        fooExpanded = lazyModule.foo + '... now with MORE!'

        # update the lazyModule with our new additions (ie, fooExpanded)
        lazyModule._updateLazyModule(globals())
    """
    class LazyLoader(object):

        """
        A data descriptor that delays instantiation of an object
        until it is first accessed.
        """

        def __init__(self, name, creator, *creatorArgs, **creatorKwargs):
            self.creator = creator
            self.args = creatorArgs
            self.kwargs = creatorKwargs
            self.name = name

        def __get__(self, obj, objtype):
            # if obj is None - ie, we're accessing this from the LazyLoadModule
            # CLASS itself (or a subclass), and not an instance of it (ie, an
            # actual module), we just return ourself - the LazyLoader object
            if obj is None:
                return self

            # In case the LazyLoader happens to get stored on more
            # than one object, cache the created object so the exact
            # same one will be returned
            if not hasattr(self, 'newobj'):
                # use the callback to create the object that will replace us
                self.newobj = self.creator(*self.args, **self.kwargs)
                if isinstance(obj, types.ModuleType) and hasattr(self.newobj, '__module__'):
                    self.newobj.__module__ = obj.__name__
            # print "Lazy-loaded object:", self.name
            # delattr( obj.__class__, self.name) # should we overwrite with None?
            # overwrite ourselves with the newly created object
            setattr(obj, self.name, self.newobj)
            return self.newobj

    # because the lazy-loaded objects need to be installed on the CLASS, not on
    # an instance, we need to ensure that all LazyLoadModules are subclasses, to
    # ensure that they aren't polluting each other's namespaces
    def __new__(cls, name, contents, autoSubClass=True):
        # because the lazy-loaded objects need to be installed on the CLASS,
        # not on an instance, we need to ensure that all LazyLoadModules are
        # subclasses, to ensure that they aren't polluting each other's
        # namespaces.  So we automatically generate a new subclass...
        if autoSubClass:
            subclassName = name.replace('.', '_') + cls.__name__
            cls = type(subclassName, (cls,), {})
        return super(LazyLoadModule, cls).__new__(cls, name)

    def __init__(self, name, contents, autoSubClass=True):
        types.ModuleType.__init__(self, name)
        self.__dict__.update(contents)
        self._lazyGlobals = contents  # globals of original module
        # add ourselves to sys.modules, overwriting the original module
        sys.modules[name] = self
        # the above line assigns a None value to all entries in the original globals.
        # luckily, we have a copy on this module we can use to restore it.
        self._lazyGlobals.update(self.__dict__)

    def __dir__(self):
        # for modules, dir usually only returns what's in the dict, and does
        # not inspect the class (ie, items on ModuleType aren't returned in
        # a module's dir, which makes sense). However, we also want to return
        # our LazyLoaded objects, to make it appear like they're there like
        # a normal object...
        keys = set(self.__dict__)
        lazyLoaders = inspect.getmembers(
            type(self), lambda x: isinstance(x, self.LazyLoader))
        keys.update(name for (name, obj) in lazyLoaders)
        return sorted(keys)

    @property
    def __all__(self):
        public = [x for x in list(self.__dict__.keys()) + list(self.__class__.__dict__.keys()) if not x.startswith('_')]
        return public

    @classmethod
    def _lazyModule_addAttr(cls, name, creator, *creatorArgs, **creatorKwargs):
        """
        Add a lazily loaded attribute to this module

        :param name: name of attribute
        :param creator: callback function to load the object
        :param creatorArgs: args passed to callback
        :param creatorKwargs: kwargs passed to callback
        :return: LazyLoader
        """
        lazyObj = cls.LazyLoader(name, creator, *creatorArgs, **creatorKwargs)
        setattr(cls, name, lazyObj)
        return lazyObj

    def __setitem__(self, attr, args):
        """
        dynModule['attrName'] = ( callbackFunc, ( 'arg1', ), {} )
        """
        # args will either be a single callable, or will be a tuple of
        # ( callable, (args,), {kwargs} )
        if hasattr(args, '__call__'):
            callback = args
        elif isinstance(args, (tuple, list)):
            if len(args) >= 1:
                assert hasattr(args[0], '__call__'), 'first argument must be callable'
                callback = args[0]
            else:
                raise ValueError("must supply at least one argument")
            if len(args) >= 2:
                assert hasattr(args[1], '__iter__'), 'second argument must be iterable'
                cb_args = args[1]
            else:
                cb_args = ()
                cb_kwargs = {}
            if len(args) == 3:
                assert isinstance(args[2], Mapping), 'third argument must be a mapping type'
                cb_kwargs = args[2]
            else:
                cb_kwargs = {}
            if len(args) > 3:
                raise ValueError(
                    "if args and kwargs are desired, they should be passed as "
                    "a tuple and dictionary, respectively")
        else:
            raise ValueError("the item must be set to a callable, or to a "
                             "3-tuple of (callable, (args,), {kwargs})")
        self._lazyModule_addAttr(attr, callback, *cb_args, **cb_kwargs)

    def __getitem__(self, attr):
        """
        return a LazyLoader without initializing it, or, if a LazyLoader does not exist with this name,
        a real object
        """
        try:
            return self.__class__.__dict__[attr]
        except KeyError:
            return self.__dict__[attr]

    # Sort of a cumbersome name, but we want to make sure it doesn't conflict with any
    # 'real' entries in the module
    def _lazyModule_update(self):
        """
        Used to update the contents of the LazyLoadModule with the contents of another dict.
        """
        self.__dict__.update(self._lazyGlobals)


# Note - since anything referencing attributes that only exist on the lazy module
# must be prefaced with a ref to the lazy module, if we are converting a pre-existing
# module to include LazyLoaded objects, we must manually go through and edit
# any references to those objects to have a 'lazyModule' prefix (or similar).
# To aid in this process, I recommend:
# 1. Uncommenting out the final print statement in _updateLazyModule
# 2. Grabbing the output of the print statement, throw it into a text editor with
#    regexp find/replace capabilities
# 3. You should have a python list of names.
#    Replace the initial and final bracket and quote - [' and '] - with opening
#    and closing parentheses - ( and )
#    Then find / replace all occurrences of:
#          ', '
#    with:
#          |
#    ...and you should be left with a regular expression you can use to find and replace
#   in your original code...
#   (you may also want to put (?<=\W) / (?=\W) in front / behind the regexp...)
#   Don't do the regexp find / replace on the source code blindly, though!
# ...also, when you make the call to _updateLazyModule that prints out the list of
# dynamic-module-only attributes, you should do it from a GUI maya - there are some objects
# that only exist in GUI-mode...

class LazyDocStringError(Exception):
    pass


class LazyDocString(object):

    """
    Set the __doc__ of an object to an instance of this class in order to have
    a docstring that is dynamically generated when used.

    Due to restrictions of inheriting from StringType (which is necessary,
    as the 'help' function does a check to see if __doc__ is a string),
    the creator can only take a single object.

    Since the object initialization requires multiple parameters, the
    LazyDocString should be given a sliceable-iterable on creation,
    of the following form:

        LazyDocString( [documentedObj, docGetter, arg1, arg2, ...] )

    documentedObj should be the object on which we are placing the docstring

    docGetter should be a function which is used to retrieve the 'real'
    docstring - it's args will be documentedObj and any extra args
    passed to the object on creation.

    Example Usage:

    >>> def getDocStringFromDict(obj):
    ...     returnVal = docStringDict[obj.__name__]
    ...     return returnVal
    >>>
    >>> # In order to alter the doc of a class, we need to use a metaclass
    >>> class TestMetaClass(type): pass
    >>>
    >>> class TestClass(with_metaclass(TestMetaClass, object)):
    ...     def aMethod(self):
    ...         pass
    ...
    ...     aMethod.__doc__ = LazyDocString( (aMethod, getDocStringFromDict, (aMethod,)) )
    >>>
    >>> TestClass.__doc__ = LazyDocString( (TestClass, getDocStringFromDict, (TestClass,)) )
    >>>
    >>>
    >>> docStringDict = {'TestClass':'New Docs for PynodeClass!',
    ...                  'aMethod':'Method docs!'}
    >>>
    >>> TestClass.__doc__
    'New Docs for PynodeClass!'
    >>> TestClass.aMethod.__doc__
    'Method docs!'


    Note that new-style classes (ie, instances of 'type') and instancemethods
    can't have their __doc__ altered.

    In the case of classes, you can get around this by using a metaclass for
    the class whose docstring you wish to alter.

    In the case of instancemethods, just set the __doc__ on the function
    underlying the method (ie, myMethod.im_func). Note that if the __doc__
    for the method is set within the class definition itself, you will
    already automatically be modifying the underlying function.
    """

    def __init__(self, argList):
        if len(argList) < 2:
            raise LazyDocStringError(
                'LazyDocString must be initialized with an iterable of the '
                'form: LazyDocString( [documentedObj, '
                'docGetter, arg1, arg2, ...] )')
        documentedObj = argList[0]
        docGetter = argList[1]
        if len(argList) > 2:
            args = argList[2]
            if len(argList) == 4:
                kwargs = argList[3]
            else:
                kwargs = {}
        else:
            args = ()
            kwargs = {}

        try:
            # put in a placeholder docstring, and check to make
            # sure we can change the __doc__ of this object!
            documentedObj.__doc__ = 'LazyDocString placeholder'
        except AttributeError:
            raise LazyDocStringError(
                'cannot modify the docstring of %r objects' %
                documentedObj.__class__.__name__)
        self.documentedObj = documentedObj
        self.docGetter = docGetter
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        # print "creating docstrings", self.docGetter, self.args, self.kwargs
        self.documentedObj.__doc__ = self.docGetter(*self.args, **self.kwargs)
        return self.documentedObj.__doc__

    def __repr__(self):
        return repr(str(self))


for _name, _method in inspect.getmembers(bytes, inspect.isroutine):
    if _name.startswith('_'):
        continue

    def makeMethod(name):
        def LazyDocStringMethodWrapper(self, *args, **kwargs):
            return getattr(str(self), name)(*args, **kwargs)
        return LazyDocStringMethodWrapper
    setattr(LazyDocString, _name, makeMethod(_name))


def addLazyDocString(obj, creator, *creatorArgs, **creatorKwargs):
    """helper for LazyDocString.  Equivalent to :

        object.__doc__ = LazyDocString( (object, creator, creatorArgs, creatorKwargs) )
    """
    lazyDoc = LazyDocString((obj, creator, creatorArgs, creatorKwargs))
    obj.__doc__ = lazyDoc


class TwoWayDict(dict):

    """
    A dictionary that can also map in reverse: value to key.

    >>> twd = TwoWayDict( {3:'foobar'} )
    >>> twd[3]
    'foobar'
    >>> twd.get_key('foobar')
    3

    Entries in both sets (keys and values) must be unique within that set, but
    not necessarily across the two sets - ie, you may have 12 as both a key and
    a value, but you may not have two keys which both map to 12 (or, as with a
    regular dict, two key entries for 12).

    If a key is updated to a new value, get_key for the old value will raise
    a KeyError:

    >>> twd = TwoWayDict( {3:'old'} )
    >>> twd[3] = 'new'
    >>> twd[3]
    'new'
    >>> twd.get_key('new')
    3
    >>> twd.get_key('old')
    Traceback (most recent call last):
        ...
    KeyError: 'old'

    Similarly, if a key is updated to an already-existing value, then the old key
    will be removed from the dictionary!

    >>> twd = TwoWayDict( {'oldKey':'aValue'} )
    >>> twd['newKey'] = 'aValue'
    >>> twd['newKey']
    'aValue'
    >>> twd.get_key('aValue')
    'newKey'
    >>> twd['oldKey']
    Traceback (most recent call last):
        ...
    KeyError: 'oldKey'

    If a group of values is fed to the TwoWayDict (either on initialization, or
    through 'update', etc) that is not consistent with these conditions, then the
    resulting dictionary is indeterminate; however, it is guaranteed to be a valid/
    uncorrupted TwoWayDict.
    (This is similar to how dict will allow, for instance, {1:'foo', 1:'bar'}).

    >>> twd = TwoWayDict( {1:'foo', 1:'bar'} )
    >>> # Is twd[1] 'foo' or 'bar'?? Nobody knows!
    >>> # ...however, one of these is guaranteed to raise an error...
    >>> twd.get_key('foo') + twd.get_key('bar')   #doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    KeyError: (either 'bar' or 'foo')
    >>> twd = TwoWayDict( {1:'foo', 2:'foo'} )
    >>> # Is twd.get_key('foo') 1 or 2? Nobody knows!
    >>> # ...however, one of these is guaranteed to raise an error...
    >>> twd[1] + twd[2]   #doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    KeyError: (either 1 or 2)

    Obviously, such shenannigans should be avoided - at some point in the future, this may
    cause an error to be raised...
    """

    def __init__(self, *args, **kwargs):
        dict.__init__(self)
        self._reverse = {}
        self.update(*args, **kwargs)

    def __setitem__(self, k, v):
        # Maintain the 1-1 mapping
        if dict.__contains__(self, k):
            del self._reverse[self[k]]
        if v in self._reverse:
            dict.__delitem__(self, self.get_key(v))
        dict.__setitem__(self, k, v)
        self._reverse[v] = k

    def has_value(self, v):
        return v in self._reverse

    def __delitem__(self, k):
        del self._reverse[self[k]]
        dict.__delitem__(self, k)

    def clear(self):
        self._reverse.clear()
        dict.clear(self)

    def copy(self):
        return TwoWayDict(self)

    def pop(self, k):
        del self._reverse[self[k]]
        return self.pop(k)

    def popitem(self, **kws):
        raise NotImplementedError()

    def setdefault(self, **kws):
        raise NotImplementedError()

    def update(self, *args, **kwargs):
        if not (args or kwargs):
            return
        if len(args) > 1:
            raise TypeError('update expected at most 1 arguments, got %d' % len(args))
        # since args may be a couple different things, cast it to a dict to
        # simplify things...
        if args:
            tempDict = dict(args[0])
        else:
            tempDict = {}
        tempDict.update(kwargs)
        for key, val in tempDict.items():
            self[key] = val

    def get_key(self, v):
        return self._reverse[v]


class EquivalencePairs(TwoWayDict):

    """
    A mapping object similar to a TwoWayDict, with the addition that indexing
    and '__contains__' can now be used with keys OR values:

    >>> eq = EquivalencePairs( {3:'foobar'} )
    >>> eq[3]
    'foobar'
    >>> eq['foobar']
    3
    >>> 3 in eq
    True
    >>> 'foobar' in eq
    True

    This is intended to be used where there is a clear distinction between
    keys and values, so there is little likelihood of the sets of keys
    and values intersecting.

    The dictionary has the same restrictions as a TwoWayDict, with the added restriction
    that an object must NOT appear in both the keys and values, unless it maps to itself.
    If a new item is set that would break this restriction, the old keys/values will be
    removed from the mapping to ensure these restrictions are met.

    >>> eq = EquivalencePairs( {1:'a', 2:'b', 3:'die'} )
    >>> eq['a']
    1
    >>> eq['b']
    2
    >>> eq[1]
    'a'
    >>> eq[2]
    'b'
    >>> del eq['die']
    >>> eq[3]
    Traceback (most recent call last):
        ...
    KeyError: 3
    >>> eq[2] = 1
    >>> eq[1]
    2
    >>> eq[2]
    1
    >>> eq['a']
    Traceback (most recent call last):
        ...
    KeyError: 'a'
    >>> eq['b']
    Traceback (most recent call last):
        ...
    KeyError: 'b'

    # Even though 2 is set as a VALUE, since it already
    # exists as a KEY, the 2:'b' mapping is removed,
    # so eq['b'] will be invalid...
    >>> eq = EquivalencePairs( {1:'a', 2:'b'} )
    >>> eq['new'] = 2
    >>> eq['new']
    2
    >>> eq[2]
    'new'
    >>> eq['b']
    Traceback (most recent call last):
        ...
    KeyError: 'b'

    # Similarly, if you set as a KEy something that
    # already exists as a value...
    >>> eq = EquivalencePairs( {1:'a', 2:'b'} )
    >>> eq['b'] = 3
    >>> eq['b']
    3
    >>> eq[3]
    'b'
    >>> eq[2]
    Traceback (most recent call last):
        ...
    KeyError: 2

    If a group of values is fed to the EquivalencePairs (either on initialization, or
    through 'update', etc) that is not consistent with it's restrictions, then the
    resulting dictionary is indeterminate; however, it is guaranteed to be a valid/
    uncorrupted TwoWayDict.

    (This is somewhat similar to the behavior of the dict object itself, which will allow
    a definition such as {1:2, 1:4} )

    Obviously, such shenannigans should be avoided - at some point in the future, this may
    even cause an error to be raised...

    Finally, note that a distinction between keys and values IS maintained, for compatibility
    with keys(), iter_values(), etc.
    """

    def __setitem__(self, k, v):
        if k in self:
            # this will check if k is in the keys OR values...
            del self[k]
        if v in self:
            del self[v]
        dict.__setitem__(self, k, v)
        self._reverse[v] = k

    def __delitem__(self, key):
        if dict.__contains__(self, key):
            super(EquivalencePairs, self).__delitem__(key)
        elif key in self._reverse:
            dict.__delitem__(self, self[key])
            del self._reverse[key]
        else:
            raise KeyError(key)

    def __getitem__(self, key):
        if dict.__contains__(self, key):
            return super(EquivalencePairs, self).__getitem__(key)
        elif key in self._reverse:
            return self._reverse[key]
        else:
            raise KeyError(key)

    def __contains__(self, key):
        return (dict.__contains__(self, key) or
                key in self._reverse)

    def get(self, key, d=None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return d


def alias(origAttrName):
    """
    Returns a property which is simply an alias for another property.

    Acts simply to provide another name to reference the same
    underlying attribute; useful when subclassing, where a subclass
    might have a more descriptive name for an attribute that has the
    same function.

    The only purpose of this function is to produce more readable code.

    Example:

    >>> class GenericExporter(object):
    ...     def __init__(self, outFile):
    ...         self.outFile = outFile
    ...
    >>> class CowExporter(GenericExporter):
    ...     cowFile = alias('outFile')
    ...
    >>> CowExporter('bessie.cow').cowFile
    'bessie.cow'
    """

    def getter(self):
        return getattr(self, origAttrName)
    getter.__name__ = "get_" + origAttrName

    def setter(self, value):
        setattr(self, origAttrName, value)

    setter.__name__ = "set_" + origAttrName
    return property(getter, setter)


class propertycache(object):

    '''Class for creating properties where the value is initially calculated then stored.

    Intended for use as a descriptor, ie:

    class MyClass(object):
        @propertycache
        def aValue(self):
            return calcValue()
    c = MyClass()
    c.aValue

    '''

    def __init__(self, func):
        self.func = func
        self.name = func.__name__

    def __get__(self, ownerInstance, ownerCls=None):
        result = self.func(ownerInstance)
        setattr(ownerInstance, self.name, result)
        return result


# unit test with doctest
if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""
.. classtree:: ProxyUni

.. dotgraph::

    main -> parse -> execute;
    main -> init;
    main -> cleanup;
    execute -> make_string;
    execute -> printf
    init -> make_string;
    main -> printf;
    execute -> compare;

"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from past.builtins import cmp
from builtins import map
from builtins import str
from past.builtins import basestring
from builtins import object
import functools
import re
import inspect
import sys
import os
import tempfile
import types
try:
    import pymel.util.external.ply.lex as lex
    import pymel.util.external.ply.yacc as yacc
except ImportError:
    import ply.lex as lex
    import ply.yacc as yacc

from .common import capitalize, uncapitalize
import warnings
from pymel.util.arguments import *
from pymel.util.utilitytypes import *


# increase from 0 to 1 or 2 for more debug feedback
def verbose():
    return 0


def currentfn():
    try:
        return sys._getframe(1).f_code.co_name
    except:
        pass


class NameParseError(Exception):
    pass


class ParsingWarning(UserWarning):
    pass


ProxyUni = proxyClass(str, 'ProxyUni', module=__name__, dataFuncName='compileName', remove=['__getitem__', '__doc__'])  # 2009 Beta 2.1 has issues with passing classes with __getitem__

# For parsed objects, Token or upper level constructs


class Parsed(ProxyUni):

    _parser = None
    _accepts = ()
    _name = None
    classes = {}

# class Parsed(unicode):
#
#    _parser = None
#    _accepts = ()

    @classmethod
    def accepts(cls, other):
        """ Checks if this Parsed class can accept another object as a subpart without reparsing """
        if isinstance(other, Parsed):
            for t in cls._accepts:
                if t == other.__class__.__name__:
                    return True
        return False

    def compileName(self):
        newname = u''
        partList = []

        def getParts(obj, newname):
            try:
                for x in obj.parts:
                    # print repr(x)
                    newname = getParts(x, newname)
            except AttributeError:
                # print "DEAD", repr(obj)
                newname += str(obj._name)
            return newname
        self._name = getParts(self, newname)
        return self._name

    @classmethod
    def getParserClass(cls, parsername):
        pass

    # init class attributes, all objects of a Parsed class share the same parser
    # TODO : check if it can be a problem with multithreading ? In that case we'll need a parser per instance
    @classmethod
    def classparserbuild(cls, **kwargs):
        """ Inits class Parser, all instances of a Parsed class share the same yacc parser object """

        clsname = cls.__name__
        try:
            # class declaration specifies a parser class
            parser = cls._parser
        except:
            # default rule
            parsername = cls.__name__ + "Parser"
            parser = Parser.classes.get(parsername, None)
            cls._parser = parser
            warnings.warn("could not read '_parser' for %s, building Parser name %s from Parsed class name %s" % (cls, parsername, clsname), UserWarning)

        if parser is not None:
            # if parser hasn't been built yet, build it
            if not isinstance(parser, Parser):
                # parser is a class
                if inspect.isclass(parser):
                    parsername = parser.__name__
                    if not issubclass(parser, Parser):
                        raise ValueError("Parser %s specified in Parsed class %s is not a Parser class" % (parsername, cls.__name__))
                # parser is a string
                elif parser in Parser.classes:
                    parsername = parser
                    parser = Parser.classes[parsername]
                else:
                    raise ValueError("Invalid Parser specification %r in Parsed class %s" % (parser, cls.__name__))

                # build class Parser, replace class _parser by the Parser instance object

                # print "Building parser instance of Parser %s for Parsed class: %s" % (parser, cls.__name__)
                # replace _parser attribute (which held a Parser class or classname) with parser class instance
                cls._parser = parser()
                cls._parser.build(**kwargs)
                # return cls._parser
        else:
            raise TypeError("Parsed class %s does not define a parser, check declarations" % cls.__name__)

    @classmethod
    def classparse(cls, data, **kwargs):
        clsname = cls.__name__
        data = str(data)
        debug = kwargs.get('debug', verbose())
        errmsg = ''
        # print "Calling parser %s with debug %s" % (cls.classparser(), debug)
        result = cls.classparser().parse(data, debug=debug)
        if cls.classparser().errorcount:
            # use error or warning ?
            errmsg = "cannot parse '%s' to a valid %s, %d parser errors" % (data, clsname, cls.classparser().errorcount)
            isValid = False
        elif not isinstance(result, cls):
            # parse successful but returned a different class than exected
            errmsg = "parsing '%s' is valid, but as a %s Parsed object, and not as a %s Parsed object as it was parsed against" % (data, result.__class__.__name__, clsname)
            isValid = False
        elif not result == data:
            # should return a Parsed object with the same string value as the parsed string
            errmsg = "parsing '%s' raised no error, but the resulting name is %s is different from the nput string %s " % (result, data)
            isValid = False
        else:
            # parse successful
            isValid = True

        # check for error in parsing and correct and raise a warning or raise an error
        # TODO : corrections and error handling
        if not isValid:
            # can try to auto-correct some badly formed names
            raise NameParseError(errmsg)

        # position is set to position of first found Parsed object
        if (result.sub):
            result._pos = result.sub[0].pos
        else:
            result._pos = 0

        result._valid = True
        return result

    @classmethod
    def classparser(cls):
        """ parser object for that class """
        return cls._parser

    # instance methods

    def parse(self, data, **kwargs):
        return self.__class__.classparse(data, **kwargs)

    @property
    def parser(self):
        """ parser object for that class """
        return self.__class__.classparser()

    @property
    def tokens(self):
        """ iterates self as leaf level lexed tokens """
        for i in expandArgs(self._sub):
            if isinstance(i, Token):
                yield i

    @property
    def sub(self):
        """ Internally stored parsing data for this Parsed object sub parts """
        return self._sub

    def setSubItem(self, index, value):
        """ Change the value of one of the Parsed sub parts.  The new value will first be parsed as the same
        type as it is replacing."""
        cls = self._sub[index].__class__
        sublist = list(self._sub)
        sublist[index] = cls(value)
        self._sub = tuple(sublist)

    @property
    def pos(self):
        """ position of that Parsed object """
        return self._pos

    def isValid(self):
        """ Validity """
        return self._valid

    def copy(self):
        """return an new independent copy of the parsed object"""
        return self.__class__(self._sub)

    def findType(self, type):
        res = []
        for x in self.sub:
            if isinstance(x, type):
                res.append(x)
            else:
                res += x.findType(type)
        return res

    def __new__(cls, *args, **kwargs):
        """ Creation of a Parsed object from a LexToken, other Parsed of compatible type or string,
            if a string is passed it will be parsed and checked for compatibility with this Parsed type """

        debug = kwargs.get('debug', verbose())
        # type checking
        data = None
        if args:
            if len(args) == 1:
                data = args[0]
            else:
                data = tuple(args)

        # some data (when initializing from single arg) can define the type of Parsed object to be created
        ptype = None
        if data is None:
            # only authorize Empty to be built without arguments
            ptype = 'Empty'
        elif isinstance(data, lex.LexToken):
            ptype = kwargs.get('type', data.type)
        elif isinstance(data, Parsed):
            ptype = data.__class__
        # can override type with the keyword 'type'
        ptype = kwargs.get('type', ptype)

        if (cls is Parsed or cls is Token):  # issubclass(cls, Token) ):
            if ptype is not None:
                if verbose():
                    print("__new__ called on Parsed/Token %s with type %r" % (cls.__name__, ptype))
                newcls = Parsed.classes.get(ptype, None)
                # can only specify an existing subclass of cls
                if newcls is None:
                    raise TypeError("Type %s does not correspond to any existing Parsed sub-class (%s does not exist)" % (ptype, cls.__name__))
                else:
                    clsname = newcls.__name__
                if not issubclass(newcls, cls):
                    raise TypeError("Type %s would create a class %s that is not a sub-class of the class %s that __new__ was called on" % (ptype, clsname, cls.__name__))
            else:
                raise TypeError("Class %s is an abstract class and can't be created directly, you must specify a valid sub-type" % (cls.__name__))
        else:
            if verbose():
                print("__new__ called on explicit class %s" % (cls.__name__))
            clsname = cls.__name__
            newcls = cls

        # print "Creating new instance of Parsed class: %r" % newcls

        # process arguments and build, check arguments compatibility with that type
        pos = None
        sub = []
        valid = False
        value = data

        # if debug : print "VALUE1", value, repr(value)

        # special case for LexToken
        if isinstance(data, lex.LexToken):
            if issubclass(newcls, Token):
                # from a unique lex Token, do not check if also creating a Token
                sub = []
                pos = data.lexpos
                value = data.value
                valid = True
            else:
                # build a Token from it
                value = Token(data.value, ptype=data.type, pos=data.pos)

        if data is None:
            # Tokens can have default value to allow initialization without arguments
            try:
                value = newcls.default()
                valid = True
            except:
                valid = False
        elif isinstance(data, newcls):
            # if debug : print "IS INSTANCE", data, repr(data)
            # from a similar class, copy it
            sub = data.sub
            pos = data.pos
            valid = data.isValid()
            value = str(data)
        elif newcls.accepts(data):
            # if debug : print "ACCEPTS", data, repr(data)
            # from a compatible Parsed sub class, build the sub list from it
            sub.append(data)
            pos = data.pos
            valid = data.isValid()
            value = str(data)
        elif isSequence(data) and not isinstance(data, basestring):
            # building from sub parts, must be of the same type and in class _accepts
            # TODO : use yacc own rules for accepts
            if data:
                valid = True
                p = 0
                for arg in data:
                    # converts LexTokens directly
                    if isinstance(arg, lex.LexToken):
                        a = Token(arg.value, ptype=arg.type, pos=data.pos)
                    else:
                        a = arg
                    # now check if it's a suitable sub-part or derived class
                    if isinstance(a, newcls):
                        sub += a.sub
                    elif newcls.accepts(a):
                        sub.append(a)
                    else:
                        valid = False
                        break
                value = u"".join(map(str, data))
                if valid:
                    pos = sub[0].pos
                else:
                    sub = []
            else:
                value = ''
        else:
            if debug:
                print("REPARSE", data, repr(data))
            # reparse unless it's a Token we already know the type of
            value = str(data)
            if issubclass(newcls, Token) and newcls is not Token:
                sub = []
                pos = 0
                valid = True
            else:
                valid = False

        # parse if necessary
        if valid:
            # print "No reparsing necessary for a resulting value %s (%r)" % (value, value)
            strvalue = str(value)
        elif isinstance(value, basestring):
            if debug:
                print("%s: Will need to reparse value %r" % (clsname, value))
            newcls.classparserbuild(debug=debug)
            if debug:
                print("VALUE", value, type(value))
            result = newcls.classparse(value, debug=debug)
            if debug:
                print("RESULT", result, type(result), isinstance(result, newcls))
            if result is not None and isinstance(result, newcls):
                strvalue = str(result)
                valid = result._valid
                sub = result._sub
                pos = result._pos
                if debug:
                    print("SUB", sub)
            else:
                strvalue = ''
                valid = False
        else:
            raise TypeError("invalid argument(s) %r, cannot be parsed to create a Parsed object of type %s" % (value, clsname))

        if valid:
            # create a unicode object with appropriate string value
            newobj = super(Parsed, cls).__new__(newcls)
            newobj._name = strvalue
            # if debug: print "NAME", newobj, type(newobj), sub#, inspect.getmro(newobj)
            # set instance attributes
            newobj._sub = tuple(sub)
            newobj._valid = valid
            # override for pos
            pos = kwargs.get('pos', pos)
            if pos is not None:
                pos += kwargs.get('offset', 0)

            if pos is None or (isinstance(pos, int) and pos >= 0):
                newobj._pos = pos
            else:
                raise ValueError("A Parsed pos can only be None or an unsigned int, %r invalid" % pos)

        return newobj

    def __add__(self, other):
        """ p1.__add__(p2) <==> p1+p2
            if p1 and p2 are of the same Parsed type, it's equivalent to reparsing str(p1) + str(p2)
            if p2 is an accepted sub part of p1, it adds it to the sub-parts
        """
        # The Parsed _accepts defines validity
        # TODO : use yacc own rules to check validity without a full reparse
        cls = self.__class__
        selfvalid = self.isValid()
        sublist = list(self.sub)
        value = str(self)
        # check other's type
        if isinstance(other, cls):
            othervalid = other.isValid()
            sublist += other.sub
        elif self.accepts(other):
            othervalid = other.isValid()
            sublist.append(other)
        elif isinstance(other, basestring):
            othervalid = False
        else:
            raise TypeError("cannot add %s and %s" % (type(self), type(other)))

        if selfvalid and othervalid:
            # no reparse
            result = cls(*sublist)
        else:
            # reparse
            result = cls(str(self) + str(other))

        return result

    def __repr__(self):
        return u"%s('%s', %s)" % (self.__class__.__name__, self, self.pos)

# Parsers, all parser must derive from the Parser class


class Parser(object):

    """ Abstract Base class for all name parsers """
    classes = {}

    def __new__(cls, *args, **kwargs):
        # this class is an abstract base class for all Parser classes, cannot be built directly

        if cls is Parser:
            # type argument can be the name of a Parser class or an instance of the class
            ptype = kwargs.get('type', None)
            if ptype is None:
                raise TypeError("must specify a Parser class")
            elif isinstance(ptype, Parser) and not ptype is Parser:
                parsercls = ptype
            elif ptype in Parser.classes:
                parsercls = Parser.classes[ptype]
            else:
                raise TypeError("invalid Parser type: %s" % ptype)
        else:
            # subclasses of Parser
            parsercls = cls

#        if 'tokensDict' in parsercls.__dict__ and 'rulesDict' in parsercls.__dict__:
#            rulesDict = parsercls.rulesDict
#            tokensDict = parsercls.tokensDict
#        else:

        reserved = []
#        #reserved = [ k for k,v in parsercls.tokensDict.items() if not k.startswith( 't_') ]
#        if hasattr( parsercls, '_reserved'):
#            for k, v in parsercls._reserved.items():
#                # reserved should be a map of reserved names (as used in the code to parse) to token names (as used in PLY).
#                # as such, we only care about the PLY names, ie. the values
#                reserved.append(v)

        rulesDict, tokensDict = Parser.getRulesAndTokens(parsercls)

        tokens = list(tokensDict.keys())
        rules = list(rulesDict.keys())
        # Sort them by line number of declaration as it's how the yacc builder works to order rules
        # TODO : some more explicit rule order handling (when precedence isn't an option) ?
        rules.sort(key=functools.cmp_to_key(
            lambda x, y: cmp(rulesDict[x].__code__.co_firstlineno,
                             rulesDict[y].__code__.co_firstlineno)))
        # print "sorted rules:", [(r, parsercls.rulesDict[r].__code__.co_firstlineno) for r in rules]
        rules = tuple(rules)

        parsercls.tokens = tuple(tokens + list(set(reserved)))
        parsercls.rules = rules

        # print "%s tokens %s" % ( parsercls, parsercls.tokens )
        # class must have a start attribute for __init__, start can be None though
        # must not inherit start as parsed would not parse own class new rules
        if not 'start' in parsercls.__dict__:
            parsercls.start = None

        # TODO : same for precedence rules
        return super(Parser, cls).__new__(parsercls, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        self.errorcount = 0
        self.lexer = None
        self.parser = None

    def t_error(self, t):
        warnings.warn("illegal character in '%s' at %i: '%s'" % (t.lexer.lexdata, t.lexpos, t.value[0]), ParsingWarning, stacklevel=1)
        self.errorcount += 1
        t.lexer.skip(1)

    def p_error(self, p):
        print("error token", p)
        if p is None:
            warnings.warn("unexpected end of file", ParsingWarning, stacklevel=1)
        else:
            warnings.warn("syntax error in '%s' at %i: '%s'" % (p.lexer.lexdata, p.lexpos, p.value), ParsingWarning, stacklevel=1)
        self.errorcount += 1

        # Just discard the token and tell the parser it's okay.
        # yacc.errok()
        # yacc.errok(). This resets the parser state so it doesn't think it's in error-recovery mode. This will prevent an error token from being generated and will reset the internal error counters so that the next syntax error will call p_error() again.
        # yacc.token(). This returns the next token on the input stream.
        # yacc.restart(). This discards the entire parsing stack and resets the parser to its initial state.

    @staticmethod
    def getRulesAndTokens(parsercls):
        """
        build the tokens and precedence tuples from inherited declarations.
        gather tokens and rules definition from Parser class members (own and inherited)
        """

        tokensDict = {}
        rulesDict = {}
        for name, obj in inspect.getmembers(parsercls):
            if name.startswith('t_') and name != 't_error':

                if isinstance(obj, basestring):
                    v = obj
                elif inspect.isfunction(obj) or inspect.ismethod(obj):
                    v = obj.__doc__
                else:
                    raise SyntaxError("Token definition %s defines neither a string nor a function, unable to parse" % m[0])
                k = name[2:]
                tokensDict[k] = obj
            # PY2: once we completely transition to PY3 only, can probably
            # remove the ismethod check
            elif name.startswith('p_') and name != 'p_error' \
                        and (inspect.isfunction(obj) or inspect.ismethod(obj)):
                k = name[2:]
                rulesDict[k] = obj
#            elif name == '_reserved' and isinstance(obj,dict):
#                # reserved attribute holds a list of reserved token keywords.
#                # these should only exist on
#                print "FOUND reserved!"
#                for k,v in obj.items():
#                    tokensDict[v]=None

        return rulesDict, tokensDict

    def build(self, **kwargs):
        debug = kwargs.get('debug', verbose())
        start = kwargs.get('start', self.__class__.start)
        parserspath = kwargs.get('outputdir', tempfile.gettempdir())
        if debug:
            print("nameparse parsers path", parserspath)
        method = kwargs.get('method', 'LALR')
        if debug:
            print("Build for", self.__class__.__name__)
            print("tokens:")
            for t in self.__class__.tokens:
                # t is not always in tokensDict, not sure if this is bad or not...
                # print "\t%s = %r" % (t, self.__class__.tokensDict[t])
                print("\t%s" % (t))
            print("rules:")
            for t in self.__class__.rules:
                # t is not always in rulesDict, not sure if this is bad or not...
                # print "\t%s = %r" % (t, self.__class__.rulesDict[t].__doc__)
                print("\t%s" % (t))
            print("start: %s" % start)

        if self.lexer is None:
            lextab = self.__class__.__name__ + "_lex"
            lkwargs = {'debug': debug, 'lextab': lextab}
            self.lexer = lex.lex(object=self, **lkwargs)
        if self.parser is None:
            tabmodule = self.__class__.__name__ + "_yacc_" + start
            pkwargs = {'outputdir': parserspath, 'debug': debug, 'tabmodule': tabmodule, 'start': start, 'method': method}
            self.parser = yacc.yacc(module=self, **pkwargs)

    def parse(self, data, **kwargs):
        self.errorcount = 0
        return self.parser.parse(data, lexer=self.lexer, **kwargs)


class Token(Parsed):

    """ A class for token types, allows direct initialization from a string and type without checking
        to avoid unnecessary double parse of the string """
    pass


# token parser, can directly use re
class TokenParser(Parser):

    """ Abstract base class for Token parser """
    _pattern = None
    _type = None

    def build(self, **kwargs):
        pattern = kwargs.get('pattern', self.__class__._pattern)
        try:
            self.parser = re.compile(pattern)
        except:
            raise ValueError("cannot build Token Parser from pattern %r. you must set the _pattern attribute to a valid regular expression" % pattern)

    def parse(self, data, **kwargs):
        self.errorcount = 0
        if self.parser.match(data) is not None:
            return Token(data, type=self._type, pos=0)
        else:
            warnings.warn("%s is not matching %s pattern %r" % (data, self.__class__.__name__, self._pattern))
            self.errorcount += 1


# special purpose empty parser
class EmptyTokenParser(Parser):

    def build(self, **kwargs):
        pass

    def parse(self, data, **kwargs):
        self.errorcount = 0
        if data:
            self.errorcount = 1
        else:
            return Empty()

# derived TokenParser classes will be built for every token definition detected in Parser classes in this module


class EmptyParser(Parser):

    """ Parser for the empty production """

    start = 'Empty'

    def p_empty(self, p):
        'Empty : '
        pass


def isParsedClass(x):
    try:
        return issubclass(x, Parsed)
    except:
        return False


def isParserClass(x):
    try:
        return issubclass(x, Parser)
    except:
        return False


def _printClassTree(cls):
    def _printTree(data, level):
        if isinstance(data, tuple):
            level += 1
            for d in data:
                _printTree(d, level)
        else:
            print(data)

    print(cls)
    tree = inspect.getclasstree(cls)
    print(tree)
    _printTree(tree)


class autoparsed(type):

    """metaclass for dramatically reducing setup syntax for simple hierarchies"""
    def __new__(mcl, classname, bases, classdict):

        newcls = super(autoparsed, mcl).__new__(mcl, classname, bases, classdict)
        module = __import__(newcls.__module__, globals(), locals(), [''])
        # print "autoparsed", newcls

        if issubclass(newcls, Parsed):
            # find the attribute _parser, which holds the parser class for this Parsed type
            parsercls = classdict.get('_parser', None)

            if isParserClass(parsercls) or (parsercls is None and '_rules' in classdict):

                rules = classdict.get('_rules', None)
                if rules is not None:
                    # there's no parser, but there is a set of rules. use this to create a default parser
                    if not isIterable(rules):
                        raise TypeError("_rules attribute must be an iterable list of string rules")

                    # first add the rules to the parserdict as attributes, these will be converted to methods, below
                    parserdict = {}
                    for i, rule in enumerate(rules):
                        attrname = 'p_%s_%02d' % (classname, i + 1)
                        parserdict[attrname] = rule
                else:
                    # parser already exists. use it, but use __dict__, not inherited info
                    parserdict = parsercls.__dict__

                # a dictionary of updates to be made to the parser
                updatedict = {}

                # automatically set the start attribute to the name of this class
                if 'start' not in parserdict:
                    updatedict['start'] = classname

                # gather up all the tokens and rules and store them separately
                tokensDict = {}
                rulesDict = {}
                reservedDict = {}
                accepts_set = set(classdict.get('_accepts', []))

                classNameReg = re.compile('[a-zA-Z][a-zA-Z0-9]*')
                for name, obj in parserdict.items():
                    # print "class %s has attribute %s" % (parsercls.__name__, m)

                    # RULES
                    if name.startswith('p_'):
                        shortname = name[2:]
                        # shorthand syntax: needs to be converted to a function
                        if isinstance(obj, basestring):
                            v = obj
                            needsWrapping = True
                        # regular syntax: already a function
                        elif inspect.isfunction(obj) or inspect.ismethod(obj):
                            v = inspect.getdoc(obj)
                            needsWrapping = False
                        else:
                            raise SyntaxError("Token definition %s defines neither a string nor a function, unable to parse" % name)

                        # process the docstring
                        accepts = []
                        colonSplit = [x.strip() for x in v.split(':')]
                        if len(colonSplit) == 1:
                            v = classname + ' : ' + v
                            colonSplit = [classname, colonSplit[0]]

                        if len(colonSplit) == 2 and colonSplit[0] != '':
                            pipeSplit = [x.strip() for x in colonSplit[1].split('|')]
                            for grp in pipeSplit:
                                # be sure to filter out ourselves
                                accepts += [x for x in grp.split() if classNameReg.match(x) and x != classname]

                        # add the classes involved in the rules
                        accepts_set.update(accepts)

                        # create the function
                        if needsWrapping:
                            def p_func(self, p):
                                p[0] = newcls(*p[1:len(p)])
                            p_func.__name__ = name
                            p_func.__doc__ = v
                            if verbose():
                                print("overwriting %s._parser.%s with yacc function: %r" % (classname, name, v))
                            updatedict[name] = p_func
                            rulesDict[name] = p_func
    #                        k = m[0][2:]
    #                        if k in tokensDict :
    #                            warnings.warn("Token %s redefined in Parser %s" % (k, parser), UserWarning)
    #                        tokensDict[k] = v
                        else:
                            rulesDict[name] = obj

                    # TOKENS
                    elif name.startswith('t_') and name != 't_error':
                        shortname = name[2:]
                        if isinstance(obj, basestring):
                            v = obj
                        elif inspect.isfunction(obj) or inspect.ismethod(obj):
                            v = obj.__doc__
                        else:
                            raise SyntaxError("Token definition %s defines neither a string nor a function, unable to parse" % name)
                        tokensDict[name] = obj

                if verbose():
                    print("%s accepts: %s" % (newcls.__name__, list(accepts_set)))

                # find bases
                bases = []
                for parsedname in accepts_set:
                    try:
                        parsedcls = getattr(module, parsedname)
                    except AttributeError:
                        warnings.warn("Could not find a Parsed class %s in module %s.  Assuming that it is an custom token." % (parsedname, module), UserWarning)
                        continue

                    try:
                        parser = parsedcls._parser
                    except AttributeError:
                        try:
                            parser = getattr(module, parsedname + 'Parser')
                        except AttributeError:
                            raise SyntaxError("Cannot find parser class for parsed class %s. It should be explicitly or automatically created as %s._parser or be named %sParser" % (parsedname, parsedname, parsedname))

                    assert isinstance(parser, Parser) or (inspect.isclass(parser) and issubclass(parser, Parser)), '%s._parser is not a subclass of Parser, it is %r' % (parsedname, parser)

                    bases.append(parser)

                # some of our bases might be subclasses of other bases.  we want the smallest list of superclasses, so weed out
                # other classes that are directly dependent on other bases
                newbases = []
                for i in bases:
                    issuperclass = False
                    #parents = inspect.getmro(i)
                    for j in bases:
                        if i in inspect.getmro(j)[1:]:
                            issuperclass = True
                            break
                    if not issuperclass:
                        newbases.append(i)

                newbases = sorted(newbases, key=lambda x: len(inspect.getmro(x)), reverse=True)

                # print "%s bases" % newcls.__name__

                for base in newbases:
                    # print base
                    # print '\t', base.tokensDict
                    # print '\t', base.rulesDict
                    tokensDict.update(base.tokensDict)
                    rulesDict.update(base.rulesDict)
#                    if hasattr(base, '_reserved'):
#                        reservedDict.update( base._reserved )

                updatedict.update(tokensDict)
                updatedict.update(rulesDict)

                # for i, parent in enumerate(inspect.getmro(base)):
                #    print '\t'*i, parent
                # print inspect.getclasstree(bases)
#                    for base in newbases:
#                        _printClassTree( base )

                if rules is not None:
                    # time to make the parser
                    # we use the Rules in the 'accepts' list to determine the bases of this parser class

                    # TODO : filter out bases that are superclasses of other bases: we only need one set

                    parserName = classname + 'Parser'
                    #newbases = sorted( newbases, key = lambda x: x.__name__)

                    updatedict['tokensDict'] = tokensDict
                    updatedict['rulesDict'] = rulesDict
#                    updatedict['_reserved'] = reservedDict

                    parserdict.update(updatedict)

                    # create parser class
                    parsercls = type(parserName, (Parser,), parserdict)
                    parsercls.__module__ = module.__name__
                    setattr(newcls, '_parser', parsercls)
                else:
                    # update the existing parser
                    for k, v in updatedict.items():
                        setattr(parsercls, k, v)

                setattr(newcls, '_accepts', tuple(accepts_set))

            elif issubclass(newcls, Token):
                # a token subclass defines its regex on the '_token' attr.
                # if this attr is defined, it causes a Parser class to be created for this token, along with
                # a rule, which is named after the class
                token = classdict.get('_token', None)
                if token is None:
                    raise TypeError("Token classes that use the autoparsed metaclass must define a _token attribute")

                # print "automatically setting up token class for %s" % newcls.__name__

                tokenBasename = classname + '_Token'
                tokenName = 't_' + tokenBasename
                ruleName = 'p_' + tokenBasename
                parserName = classname + 'Parser'

                if hasattr(token, '__name__'):
                    # print "setting token name to %s" % tokenName
                    token.__name__ = tokenName

                parserdict = {}
                parserdict[tokenName] = token

                tokensDict = {tokenName: token}

#                if hasattr(newcls, '_reserved'):
#                    print "adding _reserved to %s" % parserName
#                    parserdict['_reserved'] = newcls._reserved

                parserdict['tokensDict'] = tokensDict

                # the rule is named after the class
                doc = '''%s : %s ''' % ( classname, tokenBasename )
                #class _parser( Parser ): pass

                def rule(self, p):
                    p[0] = newcls(p[1], pos=p.lexpos(1))
                rule.__name__ = ruleName
                rule.__doc__ = doc
                parserdict[ruleName] = rule
                parserdict['rulesDict'] = {ruleName: rule}

                # create the Parser class
                print("creating parser %s" % parserName)
                parsercls = type(parserName, (Parser,), parserdict)
                parsercls.__module__ = module.__name__
                #setattr( _parser, tokenName, token )
                #setattr( _parser, ruleName, rule )
                setattr(newcls, '_parser', parsercls)

        return newcls


def _getTokenPatterns(parsercls):
    tokensDict = {}
    for name, obj in parsercls.__dict__.items():
        # print "class %s has attribute %s" % (parsercls.__name__, m)
        if name.startswith('t_') and name != 't_error':
            # strip off 't_'
            k = name[2:]
            if isinstance(obj, basestring):
                v = obj
            elif inspect.isfunction(obj) or inspect.ismethod(obj):
                v = obj.__doc__
            else:
                raise SyntaxError("Token definition %s defines neither a string nor a function, unable to parse" % name)
            k = name[2:]
            if k in tokensDict:
                warnings.warn("Token %s redefined in Parser %s" % (k, parser), UserWarning)
            tokensDict[k] = v
    return tokensDict


Parsed.classes = {}
Parser.classes = {}


def process(module=None):
    """cache out a dictionary of all Parsed and Parser classes, and create token classes"""
    if module:
        # User supplied a module object.
        if isinstance(module, types.ModuleType):
            module_dict = module.__dict__
#        elif isinstance(module, _INSTANCETYPE):
#            _items = [(k,getattr(module,k)) for k in dir(module)]
#            ldict = { }
#            for i in _items:
#                ldict[i[0]] = i[1]
        else:
            raise ValueError("Expected a module")

    else:
        # No module given.  We might be able to get information from the caller.
        # Throw an exception and unwind the traceback to get the globals

        try:
            raise RuntimeError
        except RuntimeError:
            e, b, t = sys.exc_info()
            f = t.tb_frame
            f = f.f_back           # Walk out to our calling function
            module_dict = f.f_globals    # Grab its globals dictionary

    #_createTokenClasses(ldict, debug=verbose() )

    parserDict = {}
    parsedDict = {}
    tokensDict = {}

    # gather up all tokens, parsers, and parsed
    for name, obj in module_dict.items():
        if isParserClass(obj):
            parserDict[name] = obj
            tokensDict.update(_getTokenPatterns(obj))

        # find Parsers that are stored on Parsed classes as _parser
        elif isParsedClass(obj):
            parsedDict[name] = obj
            if hasattr(obj, '_parser'):
                if isParserClass(obj._parser):
                    tokensDict.update(_getTokenPatterns(obj._parser))

    for token in tokensDict:
        pattern = tokensDict[token]
        parsedName = token
        parserName = token + "Parser"

        if verbose():
            print("adding class %s for token %s of pattern r'%s'" % (parserName, token, pattern))

        class ThisTokenParser(TokenParser):

            """ Token Parser stub class """
        # set the Token Parser class attributes
        ThisTokenParser.__name__ = parserName
        #ThisTokenParser.__doc__ = "Parser for token %s=%r" % (token, pattern)
        ThisTokenParser.__module__ = __name__
        ThisTokenParser._pattern = pattern
        ThisTokenParser._type = token
        # add to the module
        module_dict[parserName] = ThisTokenParser
        parserDict[parserName] = ThisTokenParser

        if verbose():
            print("adding class %s for token %s of pattern r'%s'" % (parsedName, token, pattern))

        class ThisToken(Token):

            """ Token stub class """
        # set the Token class attributes
        ThisToken.__name__ = parsedName
        # ThisToken.__doc__ = "Parser for token %s=%r" % (token, pattern)
        ThisToken.__module__ = __name__
        ThisToken._parser = ThisTokenParser
        # add to the module
        module_dict[parsedName] = ThisToken
        parsedDict[parsedName] = ThisToken

    Parser.classes = parserDict
    Parsed.classes = parsedDict
